/**********************************************************************
Copyright 1999 by Object Craft

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

OBJECT CRAFT DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
**********************************************************************/

#include "mssqldb.h"

PyTypeObject DateTimeType;

static void datetime_crack(DateTimeObj *self)
{
    if (self->cracked)
	return;
    dbdatecrack(NULL, &self->daterec, &self->datetime);
    self->cracked = 1;
}

static struct PyMethodDef DateTime_methods[] = {
    { NULL }			/* sentinel */
};

int datetime_as_string(PyObject *obj, char *text)
{
    int len;

    len = dbconvert(NULL,
		    SQLDATETIME, (BYTE*)&((DateTimeObj*)obj)->datetime, -1,
		    SQLCHAR, text, DATETIME_LEN);
    if (len > 0)
	text[len] = '\0';
    return len;
}

PyObject *datetime_alloc(DBDATETIME *datetime)
{
    DateTimeObj *self;

    self = PyObject_NEW(DateTimeObj, &DateTimeType);
    if (self == NULL)
	return NULL;

    memcpy(&self->datetime, datetime, sizeof(self->datetime));
    memset(&self->daterec, 0, sizeof(self->daterec));
    self->cracked = 0;
    return (PyObject*)self;
}

PyObject *DateTime_FromString(PyObject *obj)
{
    DBDATETIME datetime;
    char *str = PyString_AsString(obj);
    
    if (dbconvert(NULL,
		  SQLCHAR, str, -1,
		  SQLDATETIME, (BYTE*)&datetime, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "datetime from string conversion failed");
	return NULL;
    }
    return datetime_alloc(&datetime);
}

static void DateTime_dealloc(DateTimeObj *self)
{
    PyMem_DEL(self);
}

static int DateTime_print(DateTimeObj *self, FILE *fp, int flags)
{
    char text[DATETIME_LEN];

    datetime_as_string((PyObject*)self, text);
    fputs(text, fp);
    return 0;
}

static PyObject *DateTime_repr(DateTimeObj *self)
{
    char text[DATETIME_LEN];

    datetime_as_string((PyObject*)self, text);
    return PyString_FromString(text);
}

static PyObject *DateTime_int(DateTimeObj *v)
{
    DBINT int_value;

    if (dbconvert(NULL,
		  SQLDATETIME, (BYTE*)&v->datetime, -1,
		  SQLINT4, (BYTE*)&int_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "int conversion failed");
	return NULL;
    }
    return PyInt_FromLong(int_value);
}

static PyObject *DateTime_long(DateTimeObj *v)
{
    char *end;
    char text[DATETIME_LEN];

    datetime_as_string((PyObject*)v, text);
    return PyLong_FromString(text, &end, 10);
}

static PyObject *DateTime_float(DateTimeObj *v)
{
    DBFLT8 float_value;

    if (dbconvert(NULL,
		  SQLDATETIME, (BYTE*)&v->datetime, -1,
		  SQLFLT8, (BYTE*)&float_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "float conversion failed");
	return NULL;
    }
    return PyFloat_FromDouble(float_value);
}

static PyNumberMethods DateTime_as_number = {
    (binaryfunc)0,		/*nb_add*/
    (binaryfunc)0,		/*nb_subtract*/
    (binaryfunc)0,		/*nb_multiply*/
    (binaryfunc)0,		/*nb_divide*/
    (binaryfunc)0,		/*nb_remainder*/
    (binaryfunc)0,		/*nb_divmod*/
    (ternaryfunc)0,		/*nb_power*/
    (unaryfunc)0,		/*nb_negative*/
    (unaryfunc)0,		/*nb_positive*/
    (unaryfunc)0,		/*nb_absolute*/
    (inquiry)0,			/*nb_nonzero*/
    (unaryfunc)0,		/*nb_invert*/
    (binaryfunc)0,		/*nb_lshift*/
    (binaryfunc)0,		/*nb_rshift*/
    (binaryfunc)0,		/*nb_and*/
    (binaryfunc)0,		/*nb_xor*/
    (binaryfunc)0,		/*nb_or*/
    (coercion)0,		/*nb_coerce*/
    (unaryfunc)DateTime_int,	/*nb_int*/
    (unaryfunc)DateTime_long,	/*nb_long*/
    (unaryfunc)DateTime_float,	/*nb_float*/
    (unaryfunc)0,		/*nb_oct*/
    (unaryfunc)0,		/*nb_hex*/
};

#define OFFSET(x) offsetof(DateTimeObj, x)

#ifdef _WIN32
static struct memberlist DateTime_memberlist[] = {
    { "year",      T_INT, OFFSET(daterec.year) },
    { "quarter",   T_INT, OFFSET(daterec.quarter) },
    { "month",     T_INT, OFFSET(daterec.month) },
    { "dayofyear", T_INT, OFFSET(daterec.dayofyear) },
    { "day",       T_INT, OFFSET(daterec.day) },
    { "week",      T_INT, OFFSET(daterec.week) },
    { "weekday",   T_INT, OFFSET(daterec.weekday) },
    { "hour",      T_INT, OFFSET(daterec.hour) },
    { "minute",    T_INT, OFFSET(daterec.minute) },
    { "second",    T_INT, OFFSET(daterec.second) },
    { "millisecond", T_INT, OFFSET(daterec.millisecond) },
    { NULL }
};
#else
static struct memberlist DateTime_memberlist[] = {
    { "year",      T_INT, OFFSET(daterec.dateyear) },
    { "month",     T_INT, OFFSET(daterec.datemonth) },
    { "day",       T_INT, OFFSET(daterec.datedmonth) },
    { "dayofyear", T_INT, OFFSET(daterec.datedyear) },
    { "weekday",   T_INT, OFFSET(daterec.datedweek) },
    { "hour",      T_INT, OFFSET(daterec.datehour) },
    { "minute",    T_INT, OFFSET(daterec.dateminute) },
    { "second",    T_INT, OFFSET(daterec.datesecond) },
    { "msecond",   T_INT, OFFSET(daterec.datemsecond) },
    { "tzone",     T_INT, OFFSET(daterec.datetzone) },
    { NULL }
};
#endif

static PyObject *DateTime_getattr(DateTimeObj *self, char *name)
{
    PyObject *rv;

    datetime_crack(self);
    rv = PyMember_Get((char*)self, DateTime_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(DateTime_methods, (PyObject *)self, name);
}

static int DateTime_setattr(DateTimeObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "Cannot delete attribute");
	return -1;
    }
    return PyMember_Set((char*)self, DateTime_memberlist, name, v);
}

static char DateTimeType__doc__[] = 
"";

PyTypeObject DateTimeType = {
    PyObject_HEAD_INIT(0)
    0,				/*ob_size*/
    "DateTimeType",		/*tp_name*/
    sizeof(DateTimeObj),	/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)DateTime_dealloc,/*tp_dealloc*/
    (printfunc)DateTime_print,	/*tp_print*/
    (getattrfunc)DateTime_getattr, /*tp_getattr*/
    (setattrfunc)DateTime_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)DateTime_repr,	/*tp_repr*/
    &DateTime_as_number,	/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)0,		/*tp_str*/

    /* Space for future expansion */
    0L,0L,0L,0L,
    DateTimeType__doc__		/* Documentation string */
};

char datetime_new__doc__[] =
"datetime(s) -> DateTime\n"
"\n"
"Create a datetime object.";

/* Implement the mssqldb.datetime() method
 */
PyObject *DateTimeType_new(PyObject *module, PyObject *args)
{
    char *str;
    DBDATETIME datetime;

    if (!PyArg_ParseTuple(args, "s", &str))
	return NULL;

    if (dbconvert(NULL,
		  SQLCHAR, str, -1,
		  SQLDATETIME, (BYTE*)&datetime, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "datetime from string conversion failed");
	return NULL;
    }
    return datetime_alloc(&datetime);
}

/* Used in unpickler
 */
static PyObject *datetime_constructor = NULL;

/* Register the datetime type with the copy_reg module.  This allows
 * Python to (un)pickle datetime objects.  The equivalent Python code
 * is this:
 *
 * def pickle_datetime(n):
 *     return datetime, (str(n), n.precision, n.scale)
 * 
 * copy_reg.pickle(type(datetime(1)), pickle_datetime, datetime)
 */
char pickle_datetime__doc__[] =
"pickle_datetime(n) -> datetime, (str(n))\n"
"\n"
"Used to pickle the datetime data type.";

/* DateTime pickling function
 */
PyObject *pickle_datetime(PyObject *module, PyObject *args)
{
    DateTimeObj *obj = NULL;
    PyObject *values = NULL,
	*tuple = NULL;
    char text[DATETIME_LEN];

    if (!PyArg_ParseTuple(args, "O!", &DateTimeType, &obj))
	goto error;
    datetime_as_string((PyObject*)obj, text);
    if ((values = Py_BuildValue("(s)", text)) == NULL)
	goto error;
    tuple = Py_BuildValue("(OO)", datetime_constructor, values);

error:
    Py_XDECREF(values);
    return tuple;
}

/* Register DateTime type pickler
 */
void copy_reg_datetime(PyObject *dict)
{
    PyObject *module = NULL,
	*pickle_func = NULL,
	*pickler = NULL,
	*obj = NULL;

    module = PyImport_ImportModule("copy_reg");
    if (module == NULL)
	goto error;
    if ((pickle_func = PyObject_GetAttrString(module, "pickle")) == NULL)
	goto error;
    if ((datetime_constructor = PyDict_GetItemString(dict, "datetime")) == NULL)
	goto error;
    if ((pickler = PyDict_GetItemString(dict, "pickle_datetime")) == NULL)
	goto error;
    obj = PyObject_CallFunction(pickle_func, "OOO",
				&DateTimeType, pickler, datetime_constructor);

error:
    Py_XDECREF(obj);
    Py_XDECREF(pickle_func);
    Py_XDECREF(module);
}
