# example7.c
#
# This example illustrates the use of browse-mode routines to
# determine the source of result columns from ad hoc queries.
import sys
from mssqldb import *

import exutils

def examine_results(dbproc):
    # Determine which tables were used to generate the query results.
    tabcount = dbproc.dbtabcount()

    # dbtabcount returns -0 if the query didn't include the words FOR
    # BROWSE.
    if tabcount > 0:
        print "The following tables were used to generate these query results:"
        for tabnum in range(1, tabcount + 1):
            tabname = dbproc.dbtabname(tabnum)
            if tabname:
                if dbproc.dbtabbrowse(tabnum):
                    browsable = "browsable"
                else:
                    browsable = "not browsable"
                print "\t%s (%s)" % (tabname, browsable)
    else:
        print "Your query was not a browse-mode select."
        print "For this reason, dbtabcount()"
        print "and dbtabname() cannot be used to print"
        print "the names of the tables involved in the query."
    # Determine which tables were used to generate each result column.
    colcount = dbproc.dbnumcols()
    print "Here are the columns of the target list and their sources:"
    print "\t%-20s   %-30s   %s\n" \
          % ("Result column:", "Source:", "Browsable?")
    for colnum in range(1, colcount + 1):
        tabname, tabnum = dbproc.dbtabsource(colnum)
        sourcecolname = dbproc.dbcolsource(colnum)
        if not tabname:
            if tabcount == 0:
                fullsource = "(not browse-mode SELECT)"
            else:
                fullsource = "(result of expression)"
        else:
            fullsource = "%s.%s" % (tabname, sourcecolname)
        if dbproc.dbcolbrowse(colnum):
            browsable = "yes"
        else:
            browsable = "no"
        print "\t%-20s   %-30s   %s" \
              % (dbproc.dbcolname(colnum), fullsource, browsable)

def send_command(dbproc):
    # Allow the user to type in an ad hoc query. This query will be
    # terminated by the word "go" appearing at the beginning of the
    # line.
    #
    # If the user types the word "quit" at the beginning of a line,
    # we'll quit the program.
    print "Enter a SQL query:"
    while 1:
        cmdbuf = raw_input("> ")
        if cmdbuf == "go":
            dbproc.dbsqlexec()
            break
        elif cmdbuf == "quit":
            dbexit()
            sys.exit(0)
        else:
            # Keep reading SQL commands.
            dbproc.dbcmd(cmdbuf)

print "Demo of SQL queries in a command batch\n"

# Initialize DB-Library. */
if not dbinit():
    sys.exit(1)

# Install the user-supplied error-handling and message-handling
# routines. They are defined at the bottom of this source file.
dberrhandle(exutils.err_handler)
dbmsghandle(exutils.msg_handler)

# Allocate and initialize the LOGINREC structure to be used to open a
# connection to SQL Server.
login = dblogin()
login.dbsetluser(exutils.USER)
login.dbsetlpwd(exutils.PASSWORD)
login.dbsetlapp("example7")

dbproc = login.dbopen(exutils.SERVER)

# Allow the user to type in a series of queries. This program will be
# terminated by the word "quit", appearing at the beginning of the
# line.
while 1:
    # Send a user-generated query to SQL Server.
    send_command(dbproc)
    # Now, examine the results of any queries the user has typed in.
    command_count = 1
    while 1:
        retcode = dbproc.dbresults()
        if retcode == NO_MORE_RESULTS:
            break
        if retcode == FAIL:
            print "Command %d failed." % command_count
        else:
            if not dbproc.dbcmdrow():
                print "Command %d cannot return rows." % command_count
            else:
                # This is a command that can return rows. Let's take a
                # closer look at it.
                print "Command %d:" % command_count
                examine_results(dbproc)
                # Throw away all data rows.
                dbproc.dbcanquery()
        command_count += 1
